import org.json.*;
import traer.physics.*;
import traer.animation.*;
/// import moviemaker.*;

final float NODE_SIZE = 10;
final float EDGE_LENGTH = 20;
final float EDGE_STRENGTH = 0.2;
final float SPACER_STRENGTH = 1000;

ParticleSystem physics;
Smoother3D centroid;

Smoother2D textroid;
String textMode;
/// MovieMaker mm;
boolean movieOn;

JSONObject htmlMap;
int dateIdx;
String theDate;
String dateOrder[];
PFont dateFont;

int nodeIdx;
String nodeOrder[];
Hashtable nodePool;
int brushMask;

int newNodeTimer;

PFont followFont, staticFont;


// PROCESSING /////////////////////////////////////

void setup()
{
  size( 600, 600 );
  smooth();
  frameRate( 24 );
  strokeWeight( 2 );
  ellipseMode( CENTER );       
  
  physics = new ParticleSystem( 0, 0.25 );
  centroid = new Smoother3D( 0.92 );
  textroid = new Smoother2D( 0.996 );

  textMode = "FOLLOW"; // "STATIC" or "FOLLOW"
  followFont = loadFont("Caspari-48.vlw");
  staticFont = loadFont("Caspari-48.vlw");
    
  initialize();


  movieOn = true;
//  if (movieOn){
//      mm = new MovieMaker(this,width,height,"generations.mov", MovieMaker.JPEG, MovieMaker.HIGH,24);
//  }else{
//      print("Not recording today");
//  }


  /*
  Particle p = physics.makeParticle(1.0, 0,0,0);
  physics.makeParticle(2.0, 0,0,0);
  physics.makeParticle(3.0, 0,0,0);
  print("physics has "+physics.numberOfParticles()+" particles");
  
  for (int i=0; i<physics.numberOfParticles(); i++){
    Particle q = physics.getParticle(i);
    print(" p"+i+": "+ q.mass()+"\n");
  }

  physics.getParticle(0).kill();
  physics.tick();
  print("physics has "+physics.numberOfParticles()+" particles");
  */


  // Load in the htmlmap data
  String jsonData[] = loadStrings("tree.js");
  String jsonText = jsonData[0];
  try {
      htmlMap = new JSONObject(jsonText);

      // Store a global array of sorted datecodes
      int nextIdx = 0;
      dateOrder = new String[htmlMap.length()];
      for(Iterator e = htmlMap.keys(); e.hasNext();){
        dateOrder[nextIdx] = (String)e.next();        
        //println("date "+nextIdx+": "+dateOrder[nextIdx]);
        nextIdx++;
      }
      java.util.Arrays.sort(dateOrder);
      //for (int i=0; i<dateOrder.length; i++){ println("date "+i+": "+dateOrder[i]); }

      dateIdx=0;
      setDate(dateIdx);
      brushMask=0;
      newNodeTimer=0;
  } catch (JSONException e) {
     println (e.toString());
      //There should probably be some better exception handling here...
   }
}


void draw()
{
  physics.tick( 1.6 ); 
  if ( physics.numberOfParticles() > 1 )
    updateCentroid();
  centroid.tick();
  
  if (textMode.equals("FOLLOW")){
    textroid.tick(); 
  }
  background( #F5F1DC );
  translate( width/2 , height/2 );
  scale( centroid.z() );
  translate( -centroid.x(), -centroid.y() );
 
  drawNetwork();  
  drawHUD();

  // Add a new node after every few redraws  
  if (newNodeTimer%3==0){
    addNode();
  }
  newNodeTimer++;
  
//  if (movieOn){
//      loadPixels();
//      mm.addFrame(pixels,width,height);
//  }
  
}



void drawHUD(){

    float opacityBoost = 1.0;
    if (nodeIdx<20){
        opacityBoost = nodeIdx/20.0;
    }

    int fadeStep = nodeOrder.length - nodeIdx;
    fadeStep = min(10,fadeStep);
    float opacity = fadeStep/10.0;
    //println("opacity "+opacity);
    //fill(0,0,0,60*opacity + 90 * opacityBoost);
    fill(0,0,0,60*opacity * opacityBoost);
    textAlign(CENTER);
    noStroke();

  
    if (textMode.equals("FOLLOW")){
      float xCoord = textroid.x();
      float yCoord = textroid.y();
      textFont(followFont,48);
    
      text(prettyDate(theDate), xCoord, yCoord-20);
    }else if (textMode.equals("STATIC")){
        // Draw the date string
        pushMatrix();
        resetMatrix();  
        textFont(staticFont,48);
        text(prettyDate(theDate), width/2.0, 50);
        popMatrix();
      
    }
}

void drawNetwork()
{      
  color rootColor = color(150,20,20);
  color newColor = color(140,167,46);
  color oldColor = color(244,204,81);
  color anchorColor = color(#86AEBD);
  
  int numParticles = physics.numberOfParticles();
  int winSize = (int) max(130,numParticles);
  int winPadding = winSize - numParticles;
  if (winPadding<0){winPadding=0;}
  float relativeAge=0.0;

  // draw edges 
  stroke( #696B56 );
  strokeWeight(3);
  beginShape( LINES );
  for ( int i = 0; i < physics.numberOfSprings(); ++i )
  {
    Spring e = physics.getSpring( i );
    Particle a = e.getOneEnd();
    Particle b = e.getTheOtherEnd();
    vertex( a.position().x(), a.position().y() );
    vertex( b.position().x(), b.position().y() );
  }
  endShape();

  // draw vertices
  noStroke();
  for ( int i = 0; i < numParticles; ++i )
  {
    Particle v = physics.getParticle(i);
    color agedColor, invisColor;

    if (i==brushMask){
      // This is the root, color it red
      fill(rootColor); 
      ellipse( v.position().x(), v.position().y(), 2*NODE_SIZE, 2*NODE_SIZE );
    }else{
      // this is an old tree, interpolate color by age
      relativeAge = 1.0 - (i+winPadding) / (float)winSize;
      //println("node "+i+" has p"+relativeAge+" / "+winSize);

      agedColor = interpColor(oldColor,newColor,relativeAge);
      //invisColor = color(red(agedColor),green(agedColor),blue(agedColor),0);
      invisColor = color(153,74,15,0);
      if (relativeAge>.9){
          // This node is about to blip away, fade it appropriately
          float opacity = 1.0-(relativeAge-0.9)*10.0;
          agedColor = interpColor(agedColor,invisColor,opacity*0.7 + 0.3);
          fill(agedColor);
          ellipse( v.position().x(), v.position().y(), NODE_SIZE+(relativeAge)*6*NODE_SIZE+(1-opacity)*3*NODE_SIZE, 
                                                       NODE_SIZE+(relativeAge)*6*NODE_SIZE+(1-opacity)*3*NODE_SIZE );
      }else{      
          if (v.mass()==1.01){
              // this is an anchor node
              fill(anchorColor);
          }else{
              // this is any other node
              fill(agedColor);
          }
          ellipse( v.position().x(), v.position().y(), NODE_SIZE+(relativeAge)*6*NODE_SIZE, NODE_SIZE+(relativeAge)*6*NODE_SIZE );
      }
    }
    
  }

}

void keyPressed()
{
  if ( key == 'c' )
  {
    initialize();
    tickDate();
    return;
  }
  
  if ( key == ' ' )
  {
    addNode();
    return;
  }
}

// ME ////////////////////////////////////////////


void updateCentroid()
{
  float minWidth = 50;
  float minHeight = 50;
  float 
    xMax = Float.NEGATIVE_INFINITY, 
    xMin = Float.POSITIVE_INFINITY, 
    yMin = Float.POSITIVE_INFINITY, 
    yMax = Float.NEGATIVE_INFINITY;

  for ( int i = 0; i < physics.numberOfParticles(); ++i )
  {
    Particle p = physics.getParticle( i );
    xMax = max( xMax, p.position().x() );
    xMin = min( xMin, p.position().x() );
    yMin = min( yMin, p.position().y() );
    yMax = max( yMax, p.position().y() );
  }
  float deltaX = xMax-xMin;
  float deltaY = yMax-yMin;
  
  //println("particle xrange: "+xMin+" - "+xMax);
  //println("         yrange: "+yMin+" - "+yMax);
  //println("world size: "+deltaX+" x "+deltaY);

  // let's zoom out a little bit by exaggerating the dimensions slightly
  int increment = (int)(max(deltaX,deltaY)*0.03);
  increment = (int)max(increment, 200.0);
  
  xMin -= increment;
  yMin -= increment;
  deltaX += 2*increment;
  deltaY += 2*increment;
  
  if ( deltaY > deltaX )
    centroid.setTarget( xMin-10 + 0.5*deltaX, yMin +0.5*deltaY, height/(deltaY+50) );
  else
    centroid.setTarget( xMin-10 + 0.5*deltaX, yMin +0.5*deltaY, width/(deltaX+50) );


  // now let's handle the centroid that's trailing the root nodes
  if (physics.numberOfParticles() > brushMask){
      Particle body = physics.getParticle(brushMask);
      float xCoord = body.position().x();
      float yCoord = body.position().y();

      int xOffset = 55;
      int yOffset = 85;
      if (xCoord>textroid.x()){ xOffset *= -1; }
      if (yCoord>textroid.y()){ yOffset *= -1; }

      textroid.setXTarget(xCoord + xOffset);
      textroid.setYTarget(yCoord + yOffset);
  }

}

void addSpacersToNode( Particle p, Particle r )
{
  for ( int i = 0; i < physics.numberOfParticles(); ++i )
  {
    Particle q = physics.getParticle( i );
    if ( p != q && p != r )
      physics.makeAttraction( p, q, -SPACER_STRENGTH, 20 );
  }
}

void makeEdgeBetween( Particle a, Particle b )
{
  physics.makeSpring( a, b, EDGE_STRENGTH, EDGE_STRENGTH, EDGE_LENGTH );
}

void initialize()
{
  physics.clear();
  //physics.makeParticle();
  centroid.setValue( 0, 0, 1.0 );
  textroid.setValue( 0, 0 );
}

void addNode()
{ 
  // by default we have one node on the screen which we'll assign to the root body node
  // (which has the identifier 'body' in the nodelist for a given day
  int numParticles = physics.numberOfParticles();

  if (nodeIdx >= nodeOrder.length){
    // println("node idx "+nodeIdx+" is too big ("+nodeOrder.length+")");
    tickDate();
    return;
  }

  try{
    // find the next node to add and get all its relevant info  
    String nextName = nodeOrder[nodeIdx];
    nodeIdx++;
    JSONObject nextNode = (JSONObject)nodePool.get(nextName);
    JSONObject nextAttribs = nextNode.getJSONObject("attribs");
    String nextParent = nextAttribs.getString("parent");
    String tagName = nextAttribs.getString("tag");
    // print("Adding node: "+nextName+" ["+nextParent+"]");


    // If we're over the limit, delete an old node
    if (physics.numberOfParticles() > 130){ //nodeOrder.length*3){
      //print("dropping first node "+ physics.numberOfParticles()+" -> ");
      physics.getParticle(0).kill();
      physics.tick(0.2);
      //print(physics.numberOfParticles()+"\n");
      brushMask--;
    }

    // If this is a body tag, we're starting a new tree    
    if(tagName.equals("body")){
      if(physics.numberOfParticles()==0){
        Particle newP = physics.makeParticle();
      }else{
        // If this is the root node, it's special. It has no parent
        Particle neighborP = physics.getParticle(physics.numberOfParticles()-1);
        Particle newP = physics.makeParticle();
        newP.moveTo( neighborP.position().x() + random( -1, 1 ), 
                     neighborP.position().y() + random( -1, 1 ), 0 );
        physics.tick(0.2);
        //println("new root at i"+(physics.numberOfParticles()-1));
      }
      return;
    }
    
    // otherwise search in the current document's region
    // to find the index of the parent node
    int parentIndex = 0;
    for (int i=0; i<nodeOrder.length; i++){
      if (nodeOrder[i].equals(nextParent)){
        parentIndex = i+brushMask;
        //print("p"+(parentIndex-brushMask)+"/"+nodeOrder.length+" ");
      }
    }
    
    // Generate the new node and attach it to its parent
    //println("appending to element"+physics.numberOfParticles());
    Particle newP = physics.makeParticle();
    Particle parentP = physics.getParticle(parentIndex);
    addSpacersToNode(newP, parentP);
    makeEdgeBetween(newP, parentP);
    newP.moveTo( parentP.position().x() + random( -1, 1 ), 
                 parentP.position().y() + random( -1, 1 ), 0 );
                 
    // encode the tag type in the particle's mass (ugh, i know...)
    if (tagName.equals("a")){
        newP.setMass(1.01);
    }else if (tagName.equals("a")){
        newP.setMass(1.02);
    }
    
    
    physics.tick(0.2);
    //println("new leaf at i"+(physics.numberOfParticles()-1));
    
    
  } catch (JSONException e) {
     println (e.toString());
      //There should probably be some better exception handling here...
  }
}

void tickDate(){
  int newDateIdx = dateIdx+1;
  if (newDateIdx>=dateOrder.length){
      newDateIdx = 0;
//      if (movieOn){
//          mm.finishMovie();
//          println("Done recording");
//          movieOn = false;
//      }
  }
  setDate(newDateIdx);
  brushMask = physics.numberOfParticles();

  Particle body = physics.getParticle(brushMask-1);
  float xCoord = body.position().x();
  float yCoord = body.position().y();
  if (textMode.equals("FOLLOW")){
      textroid.setValue( xCoord, yCoord );
  }

}

void setDate(int newIdx){
 dateIdx = newIdx;
 try{
   theDate = dateOrder[newIdx];
   println("New day: "+theDate);
   JSONObject nodes = htmlMap.getJSONObject(theDate);
   nodeOrder = new String[nodes.length()];
   nodePool = new Hashtable();
   nodeIdx=0;
 
   JSONObject nextNode;
   for (Iterator e = nodes.keys(); e.hasNext();){
     String tempName = (String) e.next();
     JSONObject tempNode = nodes.getJSONObject(tempName);
     JSONObject tempAttribs = tempNode.getJSONObject("attribs");
     int tempOrder = tempAttribs.getInt("order");
     //String tempParent = tempAttribs.getInt("parent");
   
     nodePool.put(tempName,tempNode);
     nodeOrder[tempOrder] = tempName;
   }
    //for (int i=0; i<nodes.length(); i++){
    //  print(i+": "+nodeOrder[i]+"\n");
    //}

  }catch (JSONException e) {
     println (e.toString());
  }
  
}




color interpColor(color baseColor, color mixColor, float mixRatio){
  //println("mixing reds "+red(baseColor)+" and "+red(mixColor) + " to "+red(baseColor)*mixRatio + red(mixColor)*(1.0-mixRatio));
  return color(red(baseColor)*mixRatio + red(mixColor)*(1.0-mixRatio),
               green(baseColor)*mixRatio + green(mixColor)*(1.0-mixRatio),
               blue(baseColor)*mixRatio + blue(mixColor)*(1.0-mixRatio),
               alpha(baseColor)*mixRatio + alpha(mixColor)*(1.0-mixRatio));  
}

String prettyDate(String uglyDate){
 StringBuffer uglyBuffer = new StringBuffer(uglyDate);
 String year = uglyBuffer.substring(0,4);
 String month = uglyBuffer.substring(4,6);
 String day = uglyBuffer.substring(6,8);
 
 if(month.equals("01")){
  month = "January"; 
 }else if(month.equals("02")){
  month = "February"; 
 }else if(month.equals("03")){
  month = "March"; 
 }else if(month.equals("04")){
  month = "April"; 
 }else if(month.equals("05")){
  month = "May"; 
 }else if(month.equals("06")){
  month = "June"; 
 }else if(month.equals("07")){
  month = "July"; 
 }else if(month.equals("08")){
  month = "August"; 
 }else if(month.equals("09")){
  month = "September"; 
 }else if(month.equals("10")){
  month = "October"; 
 }else if(month.equals("11")){
  month = "November"; 
 }else if(month.equals("12")){
  month = "December"; 
 }
 
 if(uglyBuffer.substring(6,7).equals("0")){
   day = uglyBuffer.substring(7,8);   
 }
 
 //return day+" "+month+" "+year; 
 return month+" "+year; 
}

